#include "CppUnitLite2/CppUnitLite2.h"
#include "CppUnitLite2/TestResultStdErr.h"

#include <memory.h>
#include <iostream>
#include <assert.h>

using namespace std;

//#include "czString.h"

//
// These global variables are just for testing, to check how many allocations we have so far for strings
// :TODO: I'll need to implement Allocators for strings. This is just a temporary hack for the tests
//
int numStringTotalAllocs=0;
int numStringTotalDeallocs=0;
int numStringAllocs=0;


//
//
// Class with utility functions to manipulate strings.
// :NOTE: Don't use this directly. Use the methods provided in "BasicString" instead
//
class StringUtils
{

public:
	template<class CharType>
	static size_t getZeroTerminatedSize(const CharType *str)
	{
		const CharType *p = str;
		while (*p!=0)
			p++;
		return p - str;
	}

	template<class CharType>
	static void copy(CharType *dst, const CharType *src, size_t codePoints)
	{
		assert(dst && src);
		while(codePoints--)
		{
			*dst=*src;
			dst++;
			src++;
		}
	}

	template<class CharType>
	static bool isEqual(const CharType *a, const CharType *b, size_t codePoints)
	{
		assert(a && b);
		while (codePoints--)
		{
			if (*a!=*b)
				return false;
		}
		return true;
	}

	template<class CharType>
	static int compare(const CharType *a, const CharType *b, size_t codePoints)
	{
		assert(a && b && codePoints);
		while(codePoints--)
		{
			if (*a<*b)
				return -1;
			else if (*a>*b)
				return +1;
		}
		return 0;
	}

	template<class CharType>
	static int compareSZ(const CharType *a, const CharType *b)
	{
		assert(a && b);
		while(1)
		{
			// No need to check the end of "a", because this expression will do it.
			if (*a<*b)
				return -1;
			else if (*a>*b)
				return 1;

			if (*b==0)
				return 0;

			a++;
			b++;
		};

	}

	template<class CharType>	
	static bool isEqualSZ(const CharType *a, const CharType *b)
	{
		assert(a && b);
		while(1)
		{
			// No need to check the end of "a", because this expression will do it.
			if (*a!=*b)
				return false;

			if (*b==0)
				return true;

			a++;
			b++;
		};
	}

	template<class CharType>	
	static void set(CharType *a, size_t count, CharType val=0)
	{
		while(count--)
		{
			*a=val;
			a++;
		}
	}
};


/*
StringAllocPolicy must implement the following behavior

size_t length()
	- Must return the current number of code points

size_t capacity()
	- Must return the maximum code points the string can store without allocating more memory

CharType* getWritableDataPtr(size_t expectedLength, bool keepData)
	- Must return a pointer we can use to manipulate the string data.
	  This will be called on every non-const member function, so it can be used to implement Copy-On-Write
	- expectedLength is the size in code points the write operation will need. This means the buffer might need to grow to hold the expected length.
	- if expectedLength is 0, a NULL pointer is accepted as a return value
	- keepData tells if we should keep the current data. False means the caller doesn't need the current data, because will will fill the buffer with new data.
	  By checking for false, you can avoid wasting time copying memory around that the caller will eventually discard.

const CharType* getConstDataPtr()
	- Must return a pointer we will only use for read-only operations.
	May return NULL if the string has no data.

void setLength(size_t codePoints)
	- Should change the string length member variable, and add a zero value at the end of the string.
	  The intention is only to store the value, and null terminate the string.

*/


template<class CharType, int N> class StringCreateInPlace;

//
// Implementation of the StringAllocPolicy for shared heap.
// This implementation uses the Copy-On-Write method to share data.
//
template <class CharType>
struct StringCreateHeap
{
private:
	struct SharedData
	{
		CharType *ptr;
		size_t length; // current length of the string
		size_t maxLength; // maximum length the current block can handle
		mutable int refCounter; // counts the number of references to this string data, so that we can implement the copy on write
	};
	SharedData *m_data;

	static void* allocMem(size_t size)
	{
		numStringAllocs++;
		numStringTotalAllocs++;
		return new char[size];
	}
	static void freeMem(void *ptr)
	{
		numStringTotalDeallocs++;
		numStringAllocs--;
		delete[] (char*)ptr;
	}

	static SharedData* allocateData(size_t codePoints)
	{
		// I'm  using (codePoints+1) to allow space for the zero at the end when calling c_str()
		void *rawPtr = allocMem(sizeof(SharedData)+ ((codePoints+1)*sizeof(CharType)));
		SharedData *data = (SharedData*)rawPtr;
		data->ptr = (CharType*) ((char*)rawPtr)+sizeof(SharedData); // memory for the string data is right after the structure
		data->maxLength = codePoints;
		data->length = 0;
		data->refCounter = 1;
		return data;
	}

	bool isUnique()
	{
		return (m_data==0 || m_data->refCounter==1);
	}

public:
	void dropData()
	{
		if (m_data)
		{
			assert(m_data->refCounter>0);
			m_data->refCounter--;
			if (m_data->refCounter==0)
			{
				freeMem(m_data);
			}
			m_data = 0;
		}
	}
private:

	void grabData(SharedData *data)
	{
		m_data = data;
		if (data)
			m_data->refCounter++;
	}

	void makeUnique(size_t extraCodePoints, size_t codePointsToCopy)
	{
		assert(!isUnique());

		SharedData *newData = allocateData(m_data->maxLength+extraCodePoints);
		if (codePointsToCopy)
		{
			memcpy(newData->ptr, m_data->ptr, codePointsToCopy * sizeof(CharType));
			newData->length = codePointsToCopy;
		}
		dropData();
		m_data = newData;
	}

	void grow(size_t codePoints, bool keepData)
	{
		if (m_data==0)
		{
			m_data = allocateData(codePoints);
		}
		else
		{
			if (isUnique() || keepData==false)
			{
				SharedData *newData = allocateData(length()+codePoints);
				if (keepData)
				{
					memcpy(newData->ptr, m_data->ptr, length()*sizeof(CharType));
					newData->length = length();
				}
				dropData();
				m_data = newData;
			}
			else
			{
				makeUnique(length()+codePoints, length() );
			}
		}
	}


	//
	// Utility function to set the data from any kind of StringCreateInPlace
	//
	template<int N>
	void setAs(const StringCreateInPlace<CharType, N> &str)
	{
		dropData();

		if (str.usingHeap())
		{
			// If using the heap, then share the data
			grabData(str.m_heapString.m_data);
		}
		else
		{
			// The other string is using the quick buffer, so we need to create a copy on the heap
			StringUtils::copy(getWritableDataPtr(str.length(), false) , str.getConstDataPtr(), str.length());
			setLength(str.length());
		}
	}

public:
	explicit StringCreateHeap() : m_data(0)
	{
	}

	~StringCreateHeap()
	{
		dropData();
	}

	explicit StringCreateHeap(const StringCreateHeap<CharType> &str) 
	{
		grabData(str.m_data);
	}


	//
	// Constructor to copy from any StringCreateInPlace
	// This should help share data, in case the string is using the heap already
	//
	template<int N>
	explicit StringCreateHeap(const StringCreateInPlace<CharType, N> &str) : m_data(0)
	{
		setAs(str);
	}

	StringCreateHeap<CharType>& operator=(const StringCreateHeap<CharType> &right)
	{
		dropData();
		grabData(right.m_data);
		return *this;
	}

	//
	// Assignment operator to copy from any kind of StringCreateInPlace
	//
	template<int N>
	StringCreateHeap<CharType>& operator=(const StringCreateInPlace<CharType, N> &right)
	{
		setAs(right);
		return *this;
	}


	CharType* getWritableDataPtr(size_t expectedLength, bool keepData)
	{
		if (expectedLength==0)
		{
			if(m_data==0)
			{
				return NULL;
			}
			else if (!isUnique())
			{
				dropData();
				return NULL;
			}
		}
		else if (expectedLength>capacity())
		{
			grow(expectedLength-capacity(), keepData);
		}
		else
		{
			if (!isUnique())
				makeUnique(expectedLength-capacity(), (keepData) ? length() : 0);
		}

		return m_data->ptr;
	}

	const CharType* getConstDataPtr() const
	{
		static CharType emptyString[1];
		if (m_data)
			return m_data->ptr;
		else
			return &emptyString[0];
	}

	size_t length() const
	{
		return (m_data==0) ? 0 : m_data->length;
	}

	size_t capacity() const
	{
		return (m_data==0) ? 0 : m_data->maxLength;
	}

	void setLength(size_t codePoints)
	{
		assert(codePoints<=capacity());

		if (m_data)
		{
			m_data->length = codePoints;
			m_data->ptr[codePoints] = 0;
		}
	}

};


//
// Implementation of the StringAllocPolicy, with an in place buffer, to avoid memory allocations for small strings
// When the string data doesn't fit in the quick buffer, it reverts to a StringCreateHeap implementation
//
template <class CharType, int BufSize>
class StringCreateInPlace
{
private:
	enum
	{
		BUFFER_SIZE=BufSize
	};

	CharType m_quickBuf[BUFFER_SIZE+1]; // +1, to hold the zero at the end when calling c_str()
	size_t m_length;
	
	//
	// Utility function to set the string from any kind of StringCreateInPlace
	//
	template<int N>
	void setAs(const StringCreateInPlace<CharType, N> &str)
	{
		if (str.length()<=BUFFER_SIZE)
		{
			// Release our heap, because we'll use the quick buffer
			if (usingHeap())
				m_heapString.dropData();

			StringUtils::copy(m_quickBuf, str.getConstDataPtr(), str.length());
		}
		else
		{
			if (str.usingHeap())
				m_heapString = str.m_heapString;
			else
				StringUtils::copy(m_heapString.getWritableDataPtr(str.length(), false), str.getConstDataPtr(), str.length());
		}

		setLength(str.length());
	}

public:

	StringCreateHeap<CharType> m_heapString;

	bool usingHeap() const
	{
		return (m_heapString.capacity()>0);
	}


	~StringCreateInPlace()
	{
	}

	StringCreateInPlace()
	{
		setLength(0);
	}


	//
	// Constructor to copy from any kind of StringCreateInPlace
	// This should help share data when the string is using the heap
	//
	template<int N>
	StringCreateInPlace(const StringCreateInPlace<CharType, N> &str) : m_length(0)
	{
		setAs(str);
	}

	template<int N>
	const StringCreateInPlace<CharType,BufSize>& operator=(const StringCreateInPlace<CharType, N> &right)
	{
		setAs(right);
		return *this;
	}

	size_t length() const
	{
		return m_length;
	}

	size_t capacity() const
	{
		if (usingHeap())
			return m_heapString.capacity();
		else
			return BUFFER_SIZE;
	}

	CharType* getWritableDataPtr(size_t expectedLength, bool keepData)
	{
		if (usingHeap())
		{
			return m_heapString.getWritableDataPtr(expectedLength, keepData);
		}
		else
		{
			if (expectedLength<=BUFFER_SIZE)
			{
				return &m_quickBuf[0];
			}
			else
			{
				// Switch from the quickbuffer to a heap buffer, and copy data if necessary
				CharType *ptr = m_heapString.getWritableDataPtr(expectedLength,false);
				if (keepData)
					StringUtils::copy(ptr,m_quickBuf,m_length);
				return ptr;
			}
		}
	}

	const CharType* getConstDataPtr() const
	{
		if (usingHeap())
			return m_heapString.getConstDataPtr();
		else
			return &m_quickBuf[0];
	}

	void setLength(size_t codePoints)
	{
		assert(codePoints<= capacity());
		m_length = codePoints;
		if (usingHeap())
			m_heapString.setLength(codePoints);
		else
			m_quickBuf[codePoints] = 0;
	}

};



template <class CharType, class StringAllocPolicy>
class BasicString
{


public:

	StringAllocPolicy m_data;

	BasicString()
	{
	}

	BasicString(const CharType* str)
	{
		size_t size = StringUtils::getZeroTerminatedSize(str);
		if (size)
		{
			StringUtils::copy(m_data.getWritableDataPtr(size,false), str, size);
			m_data.setLength(size);
		}
	}

	template<class SA>
	BasicString(const BasicString<CharType, SA> &str) : m_data(str.m_data)
	{

	}

	size_t size() const
	{
		return m_data.length();
	}

	const CharType* c_str() const
	{
		return m_data.getConstDataPtr();
	}

	size_t length() const
	{
		return m_data.length();
	}

	size_t capacity() const
	{
		return m_data.capacity();
	}

	bool empty() const
	{
		return (m_data.length()==0);
	}

	void clear()
	{
		size_t len = m_data.length();
		if (len>0)
		{
			m_data.getWritableDataPtr(0,false);
			m_data.setLength(0);
		}
	}

	void reserve(size_t codePoints)
	{
		if (codePoints>m_data.capacity())
		{
			m_data.getWritableDataPtr(codePoints,true);
			m_data.setLength(length());
		}
	}

	void resize(size_t count, CharType val=0)
	{
		size_t oldSize = m_data.length();
		CharType *ptr = m_data.getWritableDataPtr(count, true);
		if (count>oldSize)
			StringUtils::set(&ptr[oldSize], count-oldSize, val);
		m_data.setLength(count);
	}


	//
	// Assignment operator to copy from any kind of Allocation policy
	//
	template<class SA>
	const BasicString<CharType, StringAllocPolicy>& operator=(const BasicString<CharType, SA> &right)
	{
		m_data = right.m_data;
		return *this;
	}

	const BasicString<CharType, StringAllocPolicy>& operator=(const CharType *right)
	{
		size_t size = StringUtils::getZeroTerminatedSize(right);
		CharType *ptr = m_data.getWritableDataPtr(size,false);
		if (size)
			StringUtils::copy(ptr, right, size);
		m_data.setLength(size);

		return *this;
	}
};


template <class C1, class SA1, class C2>
bool operator==(const BasicString<C1,SA1> &left, const C2 *right)
{
	assert(right);

	const C1 *ptr1 = left.m_data.getConstDataPtr();
	if (ptr1==right) // Points to the same memory, so it must be equal
		return true;
	else
		return StringUtils::isEqualSZ(ptr1, right);
}

template <class C1, class SA1, class C2>
bool operator!=(const BasicString<C1,SA1> &left, const C2 *right)
{
	return !(operator==(left,right));
}

template <class C1, class SA1, class C2, class SA2>
bool operator==(const BasicString<C1,SA1> &left, const BasicString<C2,SA2> &right)
{
	const C1 *ptr1 = left.m_data.getConstDataPtr();
	const C2 *ptr2 = right.m_data.getConstDataPtr();
	if (ptr1==ptr2)
		return true;
	else if (left.length()!=right.length())
		return false;
	else
		return StringUtils::isEqual(ptr1, ptr2, left.length());
}


template <class C1, class SA1, class C2, class SA2>
bool operator!=(const BasicString<C1,SA1> &left, const BasicString<C2,SA2> &right)
{
	return !(operator==(left,right));
}


template <class C1, class SA1, class C2, class SA2>
bool operator<(const BasicString<C1,SA1> &left, const BasicString<C2,SA2> &right)
{
	const size_t leftLen = left.m_data.length();
	const size_t rightLen = right.m_data.length();
	int res = StringUtils::compare(left.m_data.getConstDataPtr(), right.m_data.getConstDataPtr(), min(leftLen, rightLen));

	if (res<0 || (res==0 && leftLen<rightLen))
		return true;
	else
		return false;
}

template <class C1, class SA1, class C2, class SA2>
bool operator>(const BasicString<C1,SA1> &left, const BasicString<C2,SA2> &right)
{
	const size_t leftLen = left.m_data.length();
	const size_t rightLen = right.m_data.length();
	int res = StringUtils::compare(left.m_data.getConstDataPtr(), right.m_data.getConstDataPtr(), min(leftLen, rightLen));
	if (res>0 || (res==0 && leftLen>rightLen))
		return true;
	else
		return false;
}


typedef BasicString<char, StringCreateHeap<char>> String8;
//typedef BasicString<unsigned short, StringCreateHeap<unsigned short>> String16;
typedef BasicString<char, StringCreateInPlace<char,8>> String8_;
typedef BasicString<char, StringCreateInPlace<char,16>> String8__;


/*
TEST(TestNewString)
{
	String8 str1("Hello World!");

	String8 str2="Hello World";
	//str2 = str1;

	str2=str1;
	bool ret = str2==str2;

	CHECK(ret);

	CHECK(str2>str1);

}
*/

//using namespace cz;
TEST(TestStrings)
{
	{
		String8_ s1 = "Hello World!!!!_";		
		String8__ s2(s1);
		s2 = s1;
		String8 s3(s1);
		s3 = s1;
		s3 = s2;
	}


	{
		String8 str1;
		std::string stdStr;
		str1.clear();
		CHECK(str1=="");
		CHECK(str1.empty());
		CHECK_EQUAL(0, str1.length());
		stdStr=str1.c_str();
		CHECK(stdStr=="");


		const char *t1 = "Rui Miguel Valente Figueira";
		str1 = t1;
		CHECK(!str1.empty());
		CHECK(str1.length()==strlen(t1));
		CHECK(str1.capacity()>=strlen(t1));
		CHECK(str1==t1);
		CHECK(std::string(t1) == str1.c_str());

		//
		// Test overlapping copies against std::string, to see if the result its the same
		//
		str1 = t1;
		const char *pos1 = &str1.c_str()[4];
		str1 = pos1;
		stdStr = t1;
		const char *pos2 = &stdStr.c_str()[4];
		stdStr = pos2;
		// pos1/pos2 started as "Miguel Valente Figueira", but due to the overlapping copy, they should now be "el Valente Figueira"
		CHECK(std::string(pos1)==std::string(pos2) && std::string(pos1)=="el Valente Figueira");
		CHECK(str1=="Miguel Valente Figueira");
		CHECK(stdStr=="Miguel Valente Figueira");

		str1 = t1;
		String8 str2(str1);
		CHECK(str1==str2);
		CHECK(str1==str2.c_str());
		str1 = t1;
		CHECK(str1==str2);
		CHECK(str1==str2.c_str());

		str1 = "";
		CHECK(str1!=str2);
		CHECK(str1!=str2.c_str());
		CHECK(str1=="");
		CHECK(str1==String8(""));
		CHECK(str1.empty());
		CHECK(str1.length()==0);
		CHECK(str1.size()==0);
		CHECK(str1.capacity()>=strlen(t1));
		str1.clear();

		// Test automatic resizing of the buffer, to hold a bigger string
		String8 str3="Hello World!";
		size_t cap1 = str2.capacity();
		str3 = "Very longggggggggggggggggggggggggggggggggggggggggggggggggggg Hello World";
		CHECK(str3.capacity()>cap1);

		// Check resize and reserve
		{
			String8 str4="Hi";
			str4.resize(str4.length()+3, '!');
			CHECK(str4=="Hi!!!");
			size_t cap = str4.capacity();
			str4.reserve(cap+1); 
			CHECK(str4.capacity()>cap);
			CHECK(str4=="Hi!!!");
			CHECK(str4.length()==5);
			cout << str4.length() << endl;

			str4.resize(str4.length()-3);

			CHECK(str4=="Hi");
			CHECK(str4.length()==2);
			cout << str4.length() << endl;
		}

		// Test compare operators
		{
			String8_ s1="ABC 1 2 3 4 5 6 7 8 9 10 11 12 13";
			String8__ s2="ABC";

			CHECK(s2<s1);
			CHECK(s1>s2);
			CHECK(s1!=s2);

			String8 s3=s1;
			s2 = s1.c_str();

			CHECK(s1==s3);
			CHECK(s1==s2);
		}
	}

	cout << "Total allocs done: " << numStringTotalAllocs << endl;
	cout << "Total deallocs done: " << numStringTotalDeallocs << endl;
	cout << "Memory leaks: " << numStringAllocs << endl;

}


TEST(HelloWorldTest)
{
	int a = 2;
	CHECK_EQUAL(2,a);
}
#define LOG_ERROR 1
#define CZLOG(level, str) printf("%d-%s", level, str)

#define CZASSERT(expr)															\
	if (!(expr))																\
	{																			\
	CZLOG(LOG_ERROR, "Assert failed at :"__FILE__);		\
		assert(0 && #expr);														\
	}


#define TESTMACRO(a,b) printf(a b "_" );

int main()
{     
	TESTMACRO("Hello", __FILE__ );
	printf("\n");
	TestResultStdErr result;
	TestRegistry::Instance().Run(result);
	system("pause");
	return (result.FailureCount());
}


