//
// CrazyGaze (http://www.crazygaze.com)
// Author : Rui Figueira
// Email  : rui@crazygaze.com
// 
// --------------------------------------------------------------
//
// czString
//		Last Update : 2008/01/30
//
//

#ifndef _CZ_STRING_H_
#define _CZ_STRING_H_

#include <memory.h>
#include <assert.h>

namespace cz
{


/*
 Template class that manages string allocation on the heap.
 If you want to implement your own, you need to implement:

static void* allocate(size_t size, size_t* bytesAllocated)
	- size : Minimum of bytes to allocate
	- bytesAllocated : Used to store how many bytes were really allocated. This is usefull to allocate in blocks.
	- Return value is the pointer to the allocated memory
static void deallocate(void *ptr)
	- deallocates the "ptr" pointer.
*/

template<int BlockSize>
class DefaultStringAllocator
{
public:
	static void* allocate(size_t size, size_t* bytesAllocated)
	{
		// Round up to the nearest block
		size_t realSize = size + (BlockSize- size%BlockSize );
		*bytesAllocated = realSize;
		return new char[realSize];
	}
	static void deallocate(void* ptr)
	{
		delete[] (char*)ptr;
	}
};

//
//
// Class with utility functions to manipulate strings.
// :NOTE: Don't use this directly. Use the methods provided in "BasicString" instead
//
class StringUtils
{

public:
	template<class CharType>
	static size_t getZeroTerminatedSize(const CharType *str)
	{
		const CharType *p = str;
		while (*p!=0)
			p++;
		return p - str;
	}

	template<class CharType>
	static void copy(CharType *dst, const CharType *src, size_t codePoints)
	{
		assert(dst && src);
		while(codePoints--)
		{
			*dst=*src;
			dst++;
			src++;
		}
	}

	template<class CharType>	
	static void set(CharType *a, size_t count, CharType val=0)
	{
		while(count--)
		{
			*a=val;
			a++;
		}
	}

	//
	// :NOTE: Pay attention when specifying "codePoints" as 0, because no code points will be compared
	// and it will return true (even if the strings are different)
	//
	template<class CharType>
	static bool isEqual(const CharType *a, const CharType *b, size_t codePoints)
	{
		assert(a && b);
		while (codePoints--)
		{
			if (*a!=*b)
				return false;
			a++;
			b++;
		}
		return true;
	}

	//
	// :NOTE: Pay attention when specifying "codePoints" as 0, because no code points will be compared
	// and it will return 0 (even if the strings are different)
	//
	template<class CharType>
	static int compare(const CharType *a, const CharType *b, size_t codePoints)
	{
		assert(a && b);
		while(codePoints--)
		{
			if (*a<*b)
				return -1;
			else if (*a>*b)
				return +1;
			a++;
			b++;
		}
		return 0;
	}

};


/*
StringStoragePolicy must implement the following behavior

size_t length() const
- Must return the current number of code points

size_t capacity() const
- Must return the maximum code points the string can store without allocating more memory

CharType* getWritableDataPtr(size_t expectedCapacity, bool keepData)
- Must return a pointer we can use to manipulate the string data.
This will be called on every non-const member function, so it can be used to implement Copy-On-Write
- expectedCapacity is the size in code points the write operation will need. This means the buffer might need to grow to hold the expected length.
- if expectedLength is 0, a NULL pointer is accepted as a return value
- keepData tells if we should keep the current data. False means the caller doesn't need the current data, because will will fill the buffer with new data.
By checking for false, you can avoid wasting time copying memory around that the caller will eventually discard.

const CharType* getConstDataPtr() const
- Must return a pointer we will only use for read-only operations.
May return NULL if the string has no data.

void setLength(size_t codePoints)
- Should change the string length member variable, and add a zero value at the end of the string.
The intention is only to store the value, and null terminate the string.

*/


// Forward declaration, because StringStorageRefCounted needs it
template<class CharType, int N, class Allocator> class StringStorageQuickRefCounted;

//
// Implementation of the StringStoragePolicy for shared heap.
// This implementation uses the Copy-On-Write method to share data.
//
template <class CharType, class Allocator=DefaultStringAllocator<64>>
struct StringStorageRefCounted
{
private:
	struct SharedData
	{
		CharType *ptr;
		size_t length; // current length of the string
		size_t maxLength; // maximum length the current block can handle
		mutable int refCounter; // counts the number of references to this string data, so that we can implement the copy on write
	};
	SharedData *m_data;

	static SharedData* allocateData(size_t codePoints)
	{
		// I'm  using (codePoints+1) to allow space for the zero at the end when calling c_str()
		size_t allocatedSize;
		size_t neededSize = sizeof(SharedData)+ ((codePoints+1)*sizeof(CharType));
		void *rawPtr = Allocator::allocate(neededSize , &allocatedSize);
		SharedData *data = (SharedData*)rawPtr;
		data->ptr = (CharType*) ((char*)rawPtr)+sizeof(SharedData); // memory for the string data is right after the structure
		
		assert(((allocatedSize-neededSize)%sizeof(CharType)) == 0);

		data->maxLength = codePoints + (allocatedSize-neededSize)/sizeof(CharType);
		data->length = 0;
		data->refCounter = 1;
		return data;
	}

	bool isUnique()
	{
		return (m_data==0 || m_data->refCounter==1);
	}

public:
	void dropData()
	{
		if (m_data)
		{
			assert(m_data->refCounter>0);
			m_data->refCounter--;
			if (m_data->refCounter==0)
			{
				Allocator::deallocate(m_data);
			}
			m_data = 0;
		}
	}
private:

	void grabData(SharedData *data)
	{
		m_data = data;
		if (data)
			m_data->refCounter++;
	}

	//
	// Utility function to set the data from another StringStorageRefCounted
	//
	void setAs(const StringStorageRefCounted& str)
	{
		//
		// The best way to delay memory allocation/deallocations seems like this:
		// 1 - If we don't have a buffer, then share the data
		// 2 - If we have a buffer that we own, and it's big enough for the string, copy the data over instead of sharing it. If it doesn't fit, drop it, and share the data.
		// 3 - If we have a buffer, but we don't own it, drop it, and share the data

		if (m_data==0)
		{
			grabData(str.m_data);
		}
		else
		{
			if (isUnique() && str.length()<=capacity())
			{
				StringUtils::copy(m_data->ptr, str.getConstDataPtr(), str.length());
				setLength(str.length());
			}
			else
			{
				dropData();
				grabData(str.m_data);
			}
		}
	}

	// Utility function to set the data from any kind of StringStorageQuickRefCounted
	template<int N>
	void setAs(const StringStorageQuickRefCounted<CharType, N, Allocator> &str)
	{
		if (str.usingHeap())
		{
			// if the string is using shared data, then try to share it
			setAs(str.m_heapString);
		}
		else
		{
			// If the string is using the quick buffer, then all we can do is copy it.
			// "getWritableDataPtr" needs to be called to perform copy-on-write if necessary
			CharType *ptr = getWritableDataPtr(str.length(),false);
			if (str.length())
				StringUtils::copy(ptr , str.getConstDataPtr(), str.length());
			setLength(str.length());
		}
	}

public:

	~StringStorageRefCounted()
	{
		dropData();
	}

	//
	//
	// Constructors
	//
	//
	explicit StringStorageRefCounted() : m_data(0)
	{
	}

	// copy constructor
	explicit StringStorageRefCounted(const StringStorageRefCounted &str) 
	{
		grabData(str.m_data);
	}

	// Constructor to copy from any StringStorageQuickRefCounted
	// This should help share data, in case the string is using the heap already
	template<int N>
	explicit StringStorageRefCounted(const StringStorageQuickRefCounted<CharType, N, Allocator> &str) : m_data(0)
	{
		setAs(str);
	}

	//
	//
	// Assignment operators
	//
	//
	StringStorageRefCounted& operator=(const StringStorageRefCounted &right)
	{
		setAs(right);
		return *this;
	}

	// Assignment operator to copy from any kind of StringStorageQuickRefCounted
	template<int N>
	StringStorageRefCounted& operator=(const StringStorageQuickRefCounted<CharType, N, Allocator> &right)
	{
		setAs(right);
		return *this;
	}

	CharType* getWritableDataPtr(size_t expectedCapacity, bool keepData)
	{
		// If the expectedCapacity is zero, we can drop the shared data, and return NULL
		if (expectedCapacity==0)
		{
			if(m_data==0)
			{
				return NULL;
			}
			else if (!isUnique())
			{
				dropData();
				return NULL;
			}
		}

		if (m_data==0)
		{
			m_data = allocateData(expectedCapacity);
		}
		else if (expectedCapacity>capacity() || !isUnique())
		{
			// Create our own buffer, and drop the current one
			SharedData *newData = allocateData(expectedCapacity);
			size_t toCopy = min(expectedCapacity, length());
			newData->length = toCopy;
			if (keepData)
				memcpy(newData->ptr, m_data->ptr, toCopy * sizeof(CharType));
			dropData();
			m_data = newData;
		}
	
		return m_data->ptr;
	}

	const CharType* getConstDataPtr() const
	{
		static CharType emptyString[1];
		if (m_data)
			return m_data->ptr;
		else
			return &emptyString[0];
	}

	size_t length() const
	{
		return (m_data==0) ? 0 : m_data->length;
	}

	size_t capacity() const
	{
		if (m_data==0)
		{
			return 0;
		}
		else
		{
			assert(m_data->maxLength>0);
			return m_data->maxLength;
		}
	}

	void setLength(size_t codePoints)
	{
		assert(codePoints<=capacity());

		if (m_data)
		{
			m_data->length = codePoints;
			m_data->ptr[codePoints] = 0;
		}
	}

	//
	// Return the reference counter.
	// 0 means the string is empty with no shared data
	// 1 means the string is the only owner
	// >1 means the data is shared with other strings
	int getRefCounter()
	{
		return (m_data==0) ? 0 : m_data->refCounter;
	}

};


//
// Implementation of the StringStoragePolicy, with an in place buffer, to avoid memory allocations for small strings
// When the string data doesn't fit in the quick buffer, it reverts to a StringStorageRefCounted implementation
//
template <class CharType, int BufSize,  class Allocator=DefaultStringAllocator<64>>
class StringStorageQuickRefCounted
{
private:
	enum
	{
		BUFFER_SIZE=BufSize
	};

	CharType m_quickBuf[BUFFER_SIZE+1]; // +1, to hold the zero at the end when calling c_str()
	size_t m_length;

	// Utility function to copy from a StringStorageRefCounted
	void setAs(const StringStorageRefCounted<CharType,Allocator> &str)
	{	
		// our shared buffer, if any, must always be bigger than our quick buffer
		assert(m_heapString.capacity()==0 || m_heapString.capacity()>BUFFER_SIZE);

		// Best way to delay memory allocations/deallocations seems like:
		// 1 - If it doesn't fit in our quick buffer
		//		- Let the underlying StringStorageRefCounted take care of it
		// 2 - If it fits in our quick buffer
		//		- Use the pointer from getWriteableDataPtr, because it already performs optimally in this situation

		if (str.length()>BUFFER_SIZE) // if doesn't fit the quick buffer, then the only way is using the shared data
		{
			m_heapString = str; // Share the data. This should call operator= to increase the refcounter
		}
		else
		{
			CharType *ptr = getWritableDataPtr(str.length(),false);
			assert(ptr);
			if (str.length())
				StringUtils::copy(ptr, str.getConstDataPtr(), str.length());
		}

		setLength(str.length());
	}

	//
	// Utility function to set the string from any kind of StringStorageQuickRefCounted
	//
	template<int N>
	void setAs(const StringStorageQuickRefCounted<CharType, N, Allocator> &str)
	{
		// our shared buffer, if any, must always be bigger than our quick buffer
		assert(m_heapString.capacity()==0 || m_heapString.capacity()>BUFFER_SIZE);

		// Best way delay memory allocation/deallocation looks like:
		// 1 - If the passed string is using the shared data
		//		- Pass it to "setAs(StringStorageRefCounted)", because it should already take care of that in an optimal way.
		// 2 - If the passed string is using the quick buffer (note: this means no data sharing can occur in the steps below)
		//		- Use the pointer from getWritableDataPtr, because it already takes care of this situation

		if (str.usingHeap())
		{
			setAs(str.m_heapString);
		}
		else
		{
			if (str.length())
				StringUtils::copy(getWritableDataPtr(str.length(),false), str.getConstDataPtr(), str.length());
			setLength(str.length());
		}
	}


public:

	StringStorageRefCounted<CharType,Allocator> m_heapString;

	bool usingHeap() const
	{
		return (m_heapString.capacity()>0);
	}

	~StringStorageQuickRefCounted()
	{
	}

	//
	//
	// Constructors
	//
	//
	explicit StringStorageQuickRefCounted()
	{
		setLength(0);
	}

	// copy constructor
	explicit StringStorageQuickRefCounted(const StringStorageQuickRefCounted &str)
	{
		setAs(str);
	}

	// Constructor to copy from any kind of StringStorageQuickRefCounted
	// This should help share data when the string is using the heap
	template<int N>
	explicit StringStorageQuickRefCounted(const StringStorageQuickRefCounted<CharType, N, Allocator> &str) : m_length(0)
	{
		setAs(str);
	}

	// Constructor to copy from a heap string
	// This should help share data when the string is using the heap
	explicit StringStorageQuickRefCounted(const StringStorageRefCounted<CharType, Allocator> &str) : m_length(0)
	{
		setAs(str);
	}


	//
	//
	// Assignment operators
	//
	//
	StringStorageQuickRefCounted& operator=(const StringStorageQuickRefCounted &right)
	{
		setAs(right);
		return *this;
	}

	// Assignment operator to copy from any kind of StringStorageQuickRefCounted
	template<int N>
	StringStorageQuickRefCounted& operator=(const StringStorageQuickRefCounted<CharType, N, Allocator> &right)
	{
		setAs(right);
		return *this;
	}

	// Assignment operator to copy from a heap string
	StringStorageQuickRefCounted& operator=(const StringStorageRefCounted<CharType, Allocator> &right)
	{
		setAs(right);
		return *this;
	}


	size_t length() const
	{
		return m_length;
	}

	size_t capacity() const
	{
		// our shared buffer, if any, must always be bigger than our quick buffer
		assert(m_heapString.capacity()==0 || m_heapString.capacity()>BUFFER_SIZE);

		if (usingHeap())
			return m_heapString.capacity();
		else
			return BUFFER_SIZE;
	}


	CharType* getWritableDataPtr(size_t expectedCapacity, bool keepData)
	{
		// our shared buffer, if any, must always be bigger than our quick buffer
		assert(m_heapString.capacity()==0 || m_heapString.capacity()>BUFFER_SIZE);

		// Best way to delay memory allocations/deallocations seems like:
		// 1 - If it doesn't fit in our quick buffer
		//		- Let the underlying StringStorageRefCounted take care of it (and keep the old data if required)
		// 2 - If it fits in our quick buffer
		//		- If we don't have a shared buffer, then use the quick buffer
		//		- If we have a shared buffer
		//				- If we own it,use it.
		//				- If we don't own it, drop it, and use the quick buffer

		if (expectedCapacity>BUFFER_SIZE) // This means the only way is using the shared data
		{
			// Remember if we need to copy data from the quick buffer
			bool copyData= (keepData && !usingHeap());

			CharType* ptr= m_heapString.getWritableDataPtr(expectedCapacity, keepData);

			// Copy the old data from the quick buffer, if we're switching from the quick buffer to the shared data
			if (copyData)
				StringUtils::copy(ptr, m_quickBuf, m_length);

			return ptr;
		}
		else // Fits our quick buffer
		{
			if (m_heapString.capacity()==0) // means we don't have shared data
			{
				return &m_quickBuf[0];
			}
			else
			{
				if (m_heapString.getRefCounter()==1) // we own the shared data
				{
					assert(m_heapString.capacity()>=expectedCapacity); // Just to make sure no memory allocations/deallocations happen here
					return m_heapString.getWritableDataPtr(expectedCapacity, keepData);
				}
				else
				{
					m_heapString.dropData();
					return &m_quickBuf[0];
				}
			}
		}

	}

	const CharType* getConstDataPtr() const
	{
		if (usingHeap())
			return m_heapString.getConstDataPtr();
		else
			return &m_quickBuf[0];
	}

	void setLength(size_t codePoints)
	{
		assert(codePoints<= capacity());
		m_length = codePoints;
		if (usingHeap())
			m_heapString.setLength(codePoints);
		else
			m_quickBuf[codePoints] = 0;
	}

	//
	// Return the reference counter.
	// 0 means the string isn't using shared data. It's using the quick buffer
	// 1 means the string is using the shared data, and it's the only owner
	// >1 means the string is using shared data, and the data is shared with other strings
	int getRefCounter()
	{
		if (usingHeap())
			return (m_heapString.getRefCounter());
		else
			return 0;
	}

};



template <class CharType, class StringStoragePolicy>
class BasicString
{


public:

	StringStoragePolicy m_data;

	//
	//
	// Constructors
	//
	//
	explicit BasicString()
	{
	}
	
	// construct from a C-style string
	BasicString(const CharType* str)
	{
		size_t size = StringUtils::getZeroTerminatedSize(str);
		if (size)
		{
			StringUtils::copy(m_data.getWritableDataPtr(size,false), str, size);
			m_data.setLength(size);
		}
	}

	// copy constructor
	BasicString(const BasicString &str) : m_data(str.m_data)
	{
	}

	// construct from any kind of BasicString. The current StringStoragePolicy must also implement the respective constructors.
	template<class SA>
	BasicString(const BasicString<CharType, SA> &str) : m_data(str.m_data)
	{

	}


	//
	//
	// Assignment operators
	//
	//
	BasicString& operator=(const BasicString &right)
	{
		if ((void*)this!= (void*)&right)
			m_data = right.m_data;
		return *this;
	}

	// Assignment operator to copy from any kind of Allocation policy
	// The current StringStoragePolicy must also implement the respective assignment operators.
	template<class SA>
	BasicString& operator=(const BasicString<CharType, SA> &right)
	{
		if ((void*)this!= (void*)&right)
			m_data = right.m_data;
		return *this;
	}

	// from a C-style string
	BasicString& operator=(const CharType *right)
	{
		size_t size = StringUtils::getZeroTerminatedSize(right);
		CharType *ptr = m_data.getWritableDataPtr(size,false);
		if (size)
			StringUtils::copy(ptr, right, size);
		m_data.setLength(size);

		return *this;
	}


	size_t size() const
	{
		return m_data.length();
	}

	const CharType* c_str() const
	{
		return m_data.getConstDataPtr();
	}

	size_t length() const
	{
		return m_data.length();
	}

	size_t capacity() const
	{
		return m_data.capacity();
	}

	bool empty() const
	{
		return (m_data.length()==0);
	}

	void clear()
	{
		size_t len = m_data.length();
		if (len>0)
		{
			m_data.getWritableDataPtr(0,false);
			m_data.setLength(0);
		}
	}

	void reserve(size_t codePoints)
	{
		if (codePoints>m_data.capacity())
		{
			m_data.getWritableDataPtr(codePoints,true);
			m_data.setLength(length());
		}
	}

	void resize(size_t count, CharType val=0)
	{
		size_t oldSize = m_data.length();
		CharType *ptr = m_data.getWritableDataPtr(count, true);
		if (count>oldSize)
			StringUtils::set(&ptr[oldSize], count-oldSize, val);
		m_data.setLength(count);
	}


	template<class C2>
	int compare(const C2 *str) const
	{
		assert(str);

		const CharType *leftPtr = m_data.getConstDataPtr();
		const C2 *rightPtr = str;

		if (leftPtr==rightPtr) // same memory, so it's the same string
			return 0;

		// Compare up to the length of our own string
		size_t todo=length();
		while(todo--)
		{
			if (*leftPtr<*rightPtr)
				return -1;
			else if (*leftPtr>*rightPtr)
				return +1;
			leftPtr++;
			rightPtr++;
		}

		// :NOTE: If we reach here, it means the strings are equal up to our own length
		if (*rightPtr==0)
			return 0; // The passed string doesn't have more characters, so they are equal
		else
			return -1; // if the passed string has more characters, so our own string is "smaller"
	}

	template<class C2, class SA>
	int compare(const BasicString<C2, SA> &str) const
	{
		const size_t leftLen = this->length();
		const size_t rightLen = str.length();

		const CharType *leftPtr = m_data.getConstDataPtr();
		const C2 *rightPtr = str.m_data.getConstDataPtr();

		if (leftPtr==rightPtr) // same memory, so it's the same string
			return 0;

		int res = StringUtils::compare(leftPtr,rightPtr,min(leftLen, rightLen));

		// if the lengths are different, then a return value of 0 needs more checking, because we haven't compared all characters
		if (res==0 && leftLen!=rightLen)
			res = (leftLen<rightLen) ? -1 : +1;

		return res;
	}

	//
	//
	// append methods
	//
	//

	// to append any kind of char
	template<class C1>
	BasicString& append(const C1 &ch)
	{
		if (ch!=0)
		{
			const size_t oldLen = m_data.length();
			CharType *ptr = m_data.getWritableDataPtr(oldLen+1,true);
			ptr[oldLen] = ch;
			m_data.setLength(oldLen+1);
		}
		return *this;
	}

	// to append any kind of c-style string
	template<class C1>
	BasicString& append(const C1 *str)
	{
		const size_t appendLen = StringUtils::getZeroTerminatedSize(str);
		if (appendLen)
		{
			const size_t oldLen = m_data.length();
			CharType *ptr = m_data.getWritableDataPtr(oldLen+appendLen,true);
			StringUtils::copy(&ptr[oldLen], str, appendLen);
			m_data.setLength(oldLen+appendLen);
		}
		return *this;
	}

	// to append any kind of BasicString
	template<class C1, class SA>
	BasicString& append(const BasicString<C1,SA> &str)
	{
		const size_t appendLen = str.length();
		if (appendLen)
		{
			const size_t oldLen = m_data.length();
			CharType *ptr = m_data.getWritableDataPtr(oldLen+appendLen,true);
			StringUtils::copy(&ptr[oldLen], str.m_data.getConstDataPtr(), appendLen);
			m_data.setLength(oldLen+appendLen);
		}
		return *this;
	}

	//
	//
	// operator+
	//
	//
	// to append any kind of char
	template<class C1>
	BasicString& operator+=(const C1 &ch)
	{
		return append(ch);
	}
	// to append any kind of c-style string
	template<class C1>
	BasicString& operator+=(const C1 *str)
	{
		return append(str);
	}
	// to append any kind of BasicString
	template<class C1, class SA>
	BasicString& operator+=(const BasicString<C1,SA> &str)
	{
		return append(str);
	}


	//
	//
	// operator[]
	//
	//
	const CharType& operator[](size_t offset) const
	{
		return m_data.getConstDataPtr()[offset];
	}

	// :Warning: Because this operator returns non-const data, it allows the user to do this:
	// somestring[index] = somechar:
	//
	// This means the we need to perform copy-on-write here too, which can make this operator a bit expensive, if
	// the data is shared.
	CharType& operator[](size_t offset)
	{
		return m_data.getWritableDataPtr(m_data.length(), true)[offset];
	}

};

} // end namespace cz

//
//
// Miscellaneous binary operators
//
//

//
// operator==
//
template <class C1, class SA1, class C2>
bool operator==(const cz::BasicString<C1,SA1> &left, const C2 *right)
{
	return (left.compare(right)==0);
}

template <class C1, class SA1, class C2>
bool operator==(const C2 *left, const cz::BasicString<C1,SA1> &right)
{
	return (right.compare(left)==0);
}

template <class C1, class SA1, class C2, class SA2>
bool operator==(const cz::BasicString<C1,SA1> &left, const cz::BasicString<C2,SA2> &right)
{
	// small optimization, to avoid checking the data
	if (left.length()!=right.length())
		return false;

	return (left.compare(right)==0);
}


//
// operator!=
//
template <class C1, class SA1, class C2>
bool operator!=(const cz::BasicString<C1,SA1> &left, const C2 *right)
{
	return (left.compare(right)!=0);
}

template <class C1, class SA1, class C2>
bool operator!=(const C2 *left, const cz::BasicString<C1,SA1> &right)
{
	return (right.compare(left)!=0);
}
template <class C1, class SA1, class C2, class SA2>
bool operator!=(const cz::BasicString<C1,SA1> &left, const cz::BasicString<C2,SA2> &right)
{
	// small optimization, to avoid checking the data
	if (left.length()!=right.length())
		return true;

	return (left.compare(right)!=0);
}

//
// operator<
//
template <class C1, class SA1, class C2>
bool operator<(const cz::BasicString<C1,SA1> &left, const C2* right)
{
	return (left.compare(right)<0);
}

template <class C1, class SA1, class C2>
bool operator<(const C2 *left, const cz::BasicString<C1,SA1> &right)
{
	return (right.compare(left)>0);
}

template <class C1, class SA1, class C2, class SA2>
bool operator<(const cz::BasicString<C1,SA1> &left, const cz::BasicString<C2,SA2> &right)
{
	return (left.compare(right)<0);
}

//
// operator>
//
template <class C1, class SA1, class C2>
bool operator>(const cz::BasicString<C1,SA1> &left, const C2* right)
{
	return (left.compare(right)>0);
}

template <class C1, class SA1, class C2>
bool operator>(const C2 *left, const cz::BasicString<C1,SA1> &right)
{
	return (right.compare(left)<0);
}

template <class C1, class SA1, class C2, class SA2>
bool operator>(const cz::BasicString<C1,SA1> &left, const cz::BasicString<C2,SA2> &right)
{
	return (left.compare(right)>0);
}

//
// operator<=
//
template <class C1, class SA1, class C2>
bool operator<=(const cz::BasicString<C1,SA1> &left, const C2* right)
{
	return (left.compare(right)<=0);
}

template <class C1, class SA1, class C2>
bool operator<=(const C2 *left, const cz::BasicString<C1,SA1> &right)
{
	return (right.compare(left)>=0);
}

template <class C1, class SA1, class C2, class SA2>
bool operator<=(const cz::BasicString<C1,SA1> &left, const cz::BasicString<C2,SA2> &right)
{
	return (left.compare(right)<=0);
}

//
// operator>=
//
template <class C1, class SA1, class C2>
bool operator>=(const cz::BasicString<C1,SA1> &left, const C2* right)
{
	return (left.compare(right)>=0);
}

template <class C1, class SA1, class C2>
bool operator>=(const C2 *left, const cz::BasicString<C1,SA1> &right)
{
	return (right.compare(left)<=0);
}

template <class C1, class SA1, class C2, class SA2>
bool operator>=(const cz::BasicString<C1,SA1> &left, const cz::BasicString<C2,SA2> &right)
{
	return (left.compare(right)>=0);
}


//
//
// operator+
//
//
template <class C1, class SA1, class C2>
cz::BasicString<C1,SA1> operator+(const cz::BasicString<C1,SA1> &left, const C2* right)
{
	return cz::BasicString<C1,SA1>(left) += right;
}

template <class C1, class SA1, class C2>
cz::BasicString<C1,SA1> operator+(const C2 *left, const cz::BasicString<C1,SA1> &right)
{
	return cz::BasicString<C1,SA1>(left) += right;
}

template <class C1, class SA1, class C2, class SA2>
cz::BasicString<C1,SA1> operator+(const cz::BasicString<C1,SA1> &left, const cz::BasicString<C2,SA2> &right)
{
	return cz::BasicString<C1,SA1>(left) += right;
}

#endif
