#include "../UnitTests/CppUnitLite2/CppUnitLite2.h"
#include "../UnitTests/CppUnitLite2/TestResultStdErr.h"

#include "czString.h"
#include <iostream>
#include <memory.h>

using namespace std;

using namespace cz;

//
// Define our own allocator, just to count the number of allocations/deallocations for testing
//
int numStringTotalAllocs;
int numStringTotalDeallocs;
int numStringAllocs;
class MyStringAllocator : public DefaultStringAllocator<64>
{
public:
	static void* allocate(size_t size, size_t* bytesAllocated)
	{
		numStringAllocs++;
		numStringTotalAllocs++;
		return DefaultStringAllocator::allocate(size, bytesAllocated);
	}
	static void deallocate(void* ptr)
	{
		numStringTotalDeallocs++;
		numStringAllocs--;
		DefaultStringAllocator::deallocate(ptr);
	}
};


// Typedefs to make tests easier
typedef BasicString<char, StringStorageRefCounted<char,MyStringAllocator>> HeapString; // String using only heap
typedef BasicString<char, StringStorageQuickRefCounted<char,8,MyStringAllocator>> QuickString8; // String using a quick buffer of 8 characters, with fall back to heap
typedef BasicString<char, StringStorageQuickRefCounted<char,16,MyStringAllocator>> QuickString16;  // String using a quick buffer of 16 characters, with fall back to heap

//using namespace cz;
TEST(TestStrings)
{
	// Test constructions and assignments with different types of Strings
	{
		QuickString8 s1 = "Hello World!!!!_"; // String "s1" has a quick buffer of only 8 characters, so this should cause a memory allocation
		CHECK(numStringAllocs==1);

		QuickString16 s2(s1); // String "s2" has a quick buffer of 16 characters, so no memory allocation should occur
		CHECK(numStringAllocs==1);
		s2 = s1;
		CHECK(numStringAllocs==1);

		HeapString s3(s1); // This should reuse the shared memory of "s1" to construct s3, because "s1" it's using the heap
		CHECK(numStringAllocs==1);
		s3 = s1;
		CHECK(numStringAllocs==1);

		//  At this point "s2" is using the quickbuffer, so this should drop the current heap (which is shared with "s1"), and create a new one
		// to hold the data from "s2" 
		s3 = s2;
		CHECK(numStringTotalAllocs==2);
		CHECK(numStringAllocs==2);

		CHECK(s1==s2);
		CHECK(s2==s3);
		CHECK(s1==s3);
	}
	
	// Check for leaks
	CHECK(numStringAllocs==0);

	// Various memory management tests
	{
		const HeapString s1 = "Hello World";
		CHECK(numStringAllocs==1);
		HeapString s2(s1); // this should reuse the string data, without memory allocations
		CHECK(numStringAllocs==1);
		s2 = s1; // this should reuse the string data, without memory allocations
		CHECK(numStringAllocs==1);

		int totalAllocs = numStringTotalAllocs;
		s2 = s1.c_str(); // This should cause a new allocation to store s2
		CHECK_EQUAL(1,numStringTotalAllocs-totalAllocs);
		CHECK(numStringAllocs==2);

		CHECK(s1==s2);

		totalAllocs = numStringTotalAllocs;
		s2 = "Hello"; // s2 current buffer should handle this assignment without a memory allocation
		CHECK_EQUAL(0,numStringTotalAllocs-totalAllocs);
		CHECK(numStringAllocs==2);

		totalAllocs = numStringTotalAllocs;
		// Try assigning a string bigger than the current capacity, which should cause a memory allocation
		char* tmp = new char[s2.capacity()+2];
		memset(tmp,'0', s2.capacity()+1);
		tmp[s2.capacity()+1]=0;
		s2 = tmp;
		CHECK_EQUAL(1, numStringTotalAllocs-totalAllocs);
		CHECK(numStringAllocs==2);

		s2 = "Hi there";

		CHECK(!(s1==s2));
		CHECK(s2!=s1);
		
		CHECK(s2==s2);
		CHECK(!(s2!=s2));

	}

	// Check for leaks
	CHECK(numStringAllocs==0);

	// Test copy constructors,assignments, and memory sharing
	{
		int totalAllocs = numStringTotalAllocs;
		
		HeapString s1("Hello World!");
		CHECK(s1=="Hello World!");
		HeapString s2="Hello World!";
		CHECK(s2=="Hello World!");

		// s1 and s2 should have independent buffers
		CHECK(numStringAllocs==2);

		//  This should not cause s1 to drop its data, because priority is delaying memory allocations/deallocations
		// and the contents of s2 fit in s1 already, so, instead of dropping and sharing, it should copy.
		s1=s2;
		CHECK(numStringAllocs==2);

		CHECK(s1=="Hello World!");
		CHECK(s1==s2);

		QuickString8 s3(s1); // This should use the quick buffer, because it fits
		CHECK(numStringAllocs==2);
		s3 = s2; // This should still use the quick buffer.
		CHECK(numStringAllocs==2);
		s3 = s1; // This doesn't fit in the quick buffer, so it should share the data from s3, to delay allocations/deallocations
		CHECK(numStringAllocs==2);
		CHECK(s1.m_data.getRefCounter()==2 && s3.m_data.getRefCounter()==2);
		s3 = "Hi"; // This fits in the quick buffer, so it should drop the data that's shared with s1, and use the quick buffer
		CHECK(numStringAllocs==2);
		CHECK(s1.m_data.getRefCounter()==1 && s3.m_data.getRefCounter()==0);
		
		s3 = s1; // Share again
		CHECK(s1.m_data.getRefCounter()==2 && s3.m_data.getRefCounter()==2);
		s3 = "Hello World 2!"; // Doesn't fit in the quick buffer, so it should drop the data that's shared with s1, and allocate memory
		CHECK(s1.m_data.getRefCounter()==1 && s3.m_data.getRefCounter()==1);
		CHECK(numStringAllocs==3);

		QuickString16 s4(s3); // Quick buffer is big enough, so it should use it
		CHECK(s4.m_data.getRefCounter()==0);
		s4 = s3; // Should still use the quick buffer
		CHECK(s4.m_data.getRefCounter()==0);
		CHECK(s4==s3);

		// All the above operations should cause only this number of allocs
		CHECK_EQUAL(3, numStringTotalAllocs-totalAllocs);
	}

	// Check for leaks
	CHECK(numStringAllocs==0);


	// Test playing around with empty strings
	{
		int totalAllocs=numStringTotalAllocs;
		HeapString s1;
		s1 = "";
		CHECK(numStringAllocs==0);
		s1.clear();
		CHECK(s1=="");
		CHECK(s1.empty());
		CHECK(s1.length()==0);
		CHECK(s1==QuickString8(s1.c_str()));
		CHECK(s1.capacity()==0); // At this point, we used empty strings, so it shouldn't have allocated a buffer yet
		QuickString8 s2="";
		QuickString16 s3=s2;
		s1 = s2;
		s1 = s3.c_str();
		s2 = s1;
		s3 = s1.c_str();
		CHECK_EQUAL(0, numStringTotalAllocs-totalAllocs);

		CHECK(s1=="");
		CHECK(s2=="");
		CHECK(s3=="");
	}

	const char *a1="";
	const char *a2="Hi there";
	int res = strcmp(a1,a2);
	res = strncmp(a1,a2, 0);
	res = strncmp(a1,a2, 1);

	// Check for leaks
	CHECK(numStringAllocs==0);

	// Test compare operators
	{
		HeapString s1 ="ABC 1 2 3 4 5 6 7 8 9 10 11 12 13";
		HeapString s2 = s1;
		QuickString8 s3="ABC";
		QuickString16 s4=s2; // Should share the same memory as s1/s2
		CHECK(numStringAllocs==1);
		CHECK(s4.m_data.getRefCounter()==3);

		
		// Test assorted comparison operators between types
		CHECK(s1==s1 && s1==s1.c_str() && s1==s2 && s1!=s3 && s1==s4);
		CHECK(s3!=s1 && s3==s3 && s3!=s4 && s3==s3.c_str());
		CHECK(!(s2<s1));
		CHECK(!(s2>s1));
		CHECK(s3<s1);
		CHECK(s1>s3);
		CHECK(! (s1!=s2) );
		CHECK(! (s1!=s2.c_str()) );
		
		// Test operator ==
		CHECK(  (HeapString("") == QuickString8("")) );
		CHECK(! (HeapString("0") == QuickString8("")) );
		CHECK(! (HeapString("") == QuickString8("0")) );
		CHECK(! (HeapString("A") == QuickString8("B")) );
		CHECK(! (HeapString("B") == QuickString8("A")) );
		CHECK(  (HeapString("Hi") == QuickString8("Hi")) );
		CHECK(! (HeapString("Hi") == QuickString8("Hi There")) );
		CHECK(! (HeapString("Hi There") == QuickString8("Hi")) );
		CHECK(  (HeapString("") == "") );
		CHECK(! (HeapString("") == "Hi") );
		CHECK(! (HeapString("Hi") == "") );
		CHECK(  (HeapString("Hi") == "Hi") );
		CHECK(! (HeapString("Hi There") == "Hi") );

		// Test operator !=
		CHECK(! (HeapString("") != QuickString8("")) );
		CHECK(  (HeapString("0") != QuickString8("")) );
		CHECK(  (HeapString("") != QuickString8("0")) );
		CHECK(  (HeapString("A") != QuickString8("B")) );
		CHECK(  (HeapString("B") != QuickString8("A")) );
		CHECK(! (HeapString("Hi") != QuickString8("Hi")) );
		CHECK(  (HeapString("Hi") != QuickString8("Hi There")) );
		CHECK(  (HeapString("Hi There") != QuickString8("Hi")) );
		CHECK(! (HeapString("") != "") );
		CHECK(  (HeapString("") != "Hi") );
		CHECK(  (HeapString("Hi") != "") );
		CHECK(! (HeapString("Hi") != "Hi") );
		CHECK(  (HeapString("Hi There") != "Hi") );

		// Test operator <
		CHECK(! (HeapString("") < QuickString8("")) );
		CHECK(! (HeapString("0") < QuickString8("")) );
		CHECK(  (HeapString("") < QuickString8("0")) );
		CHECK(  (HeapString("A") < QuickString8("B")) );
		CHECK(! (HeapString("B") < QuickString8("A")) );
		CHECK(! (HeapString("Hi") < QuickString8("Hi")) );
		CHECK(  (HeapString("Hi") < QuickString8("Hi There")) );
		CHECK(! (HeapString("Hi There") < QuickString8("Hi")) );
		CHECK(! (HeapString("") < "") );
		CHECK(  (HeapString("") < "Hi") );
		CHECK(! (HeapString("Hi") < "") );
		CHECK(! (HeapString("Hi") < "Hi") );
		CHECK(! (HeapString("Hi There") < "Hi") );

		// Test operator <=
		CHECK(  (HeapString("") <= QuickString8("")) );
		CHECK(! (HeapString("0") <= QuickString8("")) );
		CHECK(  (HeapString("") <= QuickString8("0")) );
		CHECK(  (HeapString("A") <= QuickString8("B")) );
		CHECK(! (HeapString("B") <= QuickString8("A")) );
		CHECK(  (HeapString("Hi") <= QuickString8("Hi")) );
		CHECK(  (HeapString("Hi") <= QuickString8("Hi There")) );
		CHECK(! (HeapString("Hi There") <= QuickString8("Hi")) );
		CHECK(  (HeapString("") <= "") );
		CHECK(  (HeapString("") <= "Hi") );
		CHECK(! (HeapString("Hi") <= "") );
		CHECK(  (HeapString("Hi") <= "Hi") );
		CHECK(! (HeapString("Hi There") <= "Hi") );

		// Test operator >
		CHECK(! (HeapString("") > QuickString8("")) );
		CHECK(  (HeapString("0") > QuickString8("")) );
		CHECK(! (HeapString("") > QuickString8("0")) );
		CHECK(! (HeapString("A") > QuickString8("B")) );
		CHECK(  (HeapString("B") > QuickString8("A")) );
		CHECK(! (HeapString("Hi") > QuickString8("Hi")) );
		CHECK(! (HeapString("Hi") > QuickString8("Hi There")) );
		CHECK(  (HeapString("Hi There") > QuickString8("Hi")) );
		CHECK(! (HeapString("") > "") );
		CHECK(! (HeapString("") > "Hi") );
		CHECK(  (HeapString("Hi") > "") );
		CHECK(! (HeapString("Hi") > "Hi") );
		CHECK(  (HeapString("Hi There") > "Hi") );

		// Test operator >=
		CHECK(  (HeapString("") >= QuickString8("")) );
		CHECK(  (HeapString("0") >= QuickString8("")) );
		CHECK(! (HeapString("") >= QuickString8("0")) );
		CHECK(! (HeapString("A") >= QuickString8("B")) );
		CHECK(  (HeapString("B") >= QuickString8("A")) );
		CHECK(  (HeapString("Hi") >= QuickString8("Hi")) );
		CHECK(! (HeapString("Hi") >= QuickString8("Hi There")) );
		CHECK(  (HeapString("Hi There") >= QuickString8("Hi")) );
		CHECK(  (HeapString("") >= "") );
		CHECK(! (HeapString("") >= "Hi") );
		CHECK(  (HeapString("Hi") >= "") );
		CHECK(  (HeapString("Hi") >= "Hi") );
		CHECK(  (HeapString("Hi There") >= "Hi") );

		//
		// Test comparison operators, but with the left operand as a C-style string
		//
		// Test operator ==
		CHECK(  ("" == QuickString8("")) );
		CHECK(! ("0" == QuickString8("")) );
		CHECK(! ("" == QuickString8("0")) );
		CHECK(! ("A" == QuickString8("B")) );
		CHECK(! ("B" == QuickString8("A")) );
		CHECK(  ("Hi" == QuickString8("Hi")) );
		CHECK(! ("Hi" == QuickString8("Hi There")) );
		CHECK(! ("Hi There" == QuickString8("Hi")) );

		// Test operator !=
		CHECK(! ("" != QuickString8("")) );
		CHECK(  ("0" != QuickString8("")) );
		CHECK(  ("" != QuickString8("0")) );
		CHECK(  ("A" != QuickString8("B")) );
		CHECK(  ("B" != QuickString8("A")) );
		CHECK(! ("Hi" != QuickString8("Hi")) );
		CHECK(  ("Hi" != QuickString8("Hi There")) );
		CHECK(  ("Hi There" != QuickString8("Hi")) );

		// Test operator <
		CHECK(! ("" < QuickString8("")) );
		CHECK(! ("0" < QuickString8("")) );
		CHECK(  ("" < QuickString8("0")) );
		CHECK(  ("A" < QuickString8("B")) );
		CHECK(! ("B" < QuickString8("A")) );
		CHECK(! ("Hi" < QuickString8("Hi")) );
		CHECK(  ("Hi" < QuickString8("Hi There")) );
		CHECK(! ("Hi There" < QuickString8("Hi")) );

		// Test operator <=
		CHECK(  ("" <= QuickString8("")) );
		CHECK(! ("0" <= QuickString8("")) );
		CHECK(  ("" <= QuickString8("0")) );
		CHECK(  ("A" <= QuickString8("B")) );
		CHECK(! ("B" <= QuickString8("A")) );
		CHECK(  ("Hi" <= QuickString8("Hi")) );
		CHECK(  ("Hi" <= QuickString8("Hi There")) );
		CHECK(! ("Hi There" <= QuickString8("Hi")) );

		// Test operator >
		CHECK(! ("" > QuickString8("")) );
		CHECK(  ("0" > QuickString8("")) );
		CHECK(! ("" > QuickString8("0")) );
		CHECK(! ("A" > QuickString8("B")) );
		CHECK(  ("B" > QuickString8("A")) );
		CHECK(! ("Hi" > QuickString8("Hi")) );
		CHECK(! ("Hi" > QuickString8("Hi There")) );
		CHECK(  ("Hi There" > QuickString8("Hi")) );

		// Test operator >=
		CHECK(  ("" >= QuickString8("")) );
		CHECK(  ("0" >= QuickString8("")) );
		CHECK(! ("" >= QuickString8("0")) );
		CHECK(! ("A" >= QuickString8("B")) );
		CHECK(  ("B" >= QuickString8("A")) );
		CHECK(  ("Hi" >= QuickString8("Hi")) );
		CHECK(! ("Hi" >= QuickString8("Hi There")) );
		CHECK(  ("Hi There" >= QuickString8("Hi")) );

	}

	// Check for leaks
	CHECK(numStringAllocs==0);

	// Test append with independent string
	{
		QuickString8 s1("Hello");
		s1 += ' ';
		s1 += HeapString("W");
		s1 += QuickString16("orld!");
		CHECK(s1=="Hello World!");
		CHECK(s1.length()==strlen("Hello World!"));
		CHECK(numStringAllocs==1);
	}


	// Test append with same string
	{
		HeapString s1("123");
		s1 += "4";
		CHECK(s1=="1234");
		s1 += s1.c_str();
		CHECK(s1=="12341234");
		s1 += s1;
		CHECK(s1=="1234123412341234");
		CHECK(s1.length()==strlen("1234123412341234"));
	}

	// Test operator+
	{
		HeapString s1="Hello";
		s1 = s1 + " ";
		s1 +=  "W" + QuickString8("o");
		s1 +=  QuickString8("r") + "l";
		s1 += QuickString8("d") + HeapString("!");
		CHECK(s1=="Hello World!");
	}
	CHECK(numStringAllocs==0);

	// Test operator[]
	{
		HeapString s1="Hello";
		const QuickString8 s2="HELLO";
		s1[1] = s2[1];
		CHECK(s1=="HEllo");
	}

	CHECK(numStringAllocs==0);

	cout << endl;
	cout << "Total allocs done: " << numStringTotalAllocs << endl;
	cout << "Total deallocs done: " << numStringTotalDeallocs << endl;
	cout << "Memory leaks: " << numStringAllocs << endl;
	cout << endl;
}


int main()
{     
	printf("\n");
	TestResultStdErr result;
	TestRegistry::Instance().Run(result);
	system("pause");
	return (result.FailureCount());
}


